------------------------------------------------------------------------------
--
--  Xilinx, Inc. 2002                 www.xilinx.com
--
--  XAPP 291 - Self Addressing FIFO
--
------------------------------------------------------------------------------
--
--  File name :       saf_1to2.vhd
--
--  Description :     Top-level for 1:2 SAF
--		      addrsize used to define the fifo depth(ie number of address bits), minimum 4,
--		      maximum 10 with ram in 1024x18/512x36 mode
-- 		      datasize defines the input data path width, minimum 1, maximum 14
-- 		      addrsize + datasize must be less than or equal 18
-- 		      addrsize + datasize must be less than or equal 36
-- 		      This example uses addrsize = 10, and datasize = 9
-- 		      flag(0) high when fifo between empty and one half full
-- 		      flag(1) high when fifo between one quarter and three quarters full
-- 		      flag(2) high when fifo between one half full and full
--
--  Date - revision : December 11th 2002 - v 1.2
--
--  Author :          NJS
--
--  Contact : e-mail  hotline@xilinx.com
--            phone   + 1 800 255 7778
--
--  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
--              provided to you "as is". Xilinx and its licensors make and you
--              receive no warranties or conditions, express, implied,
--              statutory or otherwise, and Xilinx specifically disclaims any
--              implied warranties of merchantability, non-infringement,or
--              fitness for a particular purpose. Xilinx does not warrant that
--              the functions contained in these designs will meet your
--              requirements, or that the operation of these designs will be
--              uninterrupted or error free, or that defects in the Designs
--              will be corrected. Furthermore, Xilinx does not warrantor
--              make any representations regarding use or the results of the
--              use of the designs in terms of correctness, accuracy,
--              reliability, or otherwise.
--
--              LIMITATION OF LIABILITY. In no event will Xilinx or its
--              licensors be liable for any loss of data, lost profits,cost
--              or procurement of substitute goods or services, or for any
--              special, incidental, consequential, or indirect damages
--              arising from the use or operation of the designs or
--              accompanying documentation, however caused and on any theory
--              of liability. This limitation will apply even if Xilinx
--              has been advised of the possibility of such damage. This
--              limitation shall apply not-withstanding the failure of the
--              essential purpose of any limited remedies herein.
--
--  Copyright  2002 Xilinx, Inc.
--  All rights reserved
--
------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

--pragma translate_off
library unisim ;
use unisim.vcomponents.all ;
--pragma translate_on

entity saf_1to2 is
generic (
	addrsize 	: integer  :=10 ;
	datasize 	: integer  :=8 );
port (
	wr_clk 		: in std_logic;
	rd_clk 		: in std_logic;
	datain 		: in std_logic_vector((datasize-1) downto 0);
	rst 		: in std_logic;
	we 		: in std_logic;
	oe		: in std_logic ;
	flag		: out std_logic_vector(2 downto 0) ;
	dataout		: out std_logic_vector(((2*datasize)-1) downto 0));
end saf_1to2;

architecture arch_saf_1to2 of saf_1to2 is

component ramb16_s18_s36 port (
	DIA    		: in std_logic_vector(15 downto 0);
	DIB    		: in std_logic_vector(31 downto 0);
	DIPA    	: in std_logic_vector(1 downto 0);
	DIPB    	: in std_logic_vector(3 downto 0);
	ENA    		: in std_logic;
	ENB    		: in std_logic;
	WEA    		: in std_logic;
	WEB    		: in std_logic;
	SSRA   		: in std_logic;
	SSRB   		: in std_logic;
	clkA   		: in std_logic;
	clkB   		: in std_logic;
	ADDRA  		: in std_logic_vector(9 downto 0);
	ADDRB  		: in std_logic_vector(8 downto 0);
	DOA    		: out std_logic_vector(15 downto 0);
	DOB    		: out std_logic_vector(31 downto 0);
	DOPA    	: out std_logic_vector(1 downto 0);
	DOPB    	: out std_logic_vector(3 downto 0));
end component ;

component incrementer
generic (
	size 		: integer) ;
port (
	ip 		: in std_logic_vector (addrsize-1 downto 0);
	op 		: out std_logic_vector(addrsize-1 downto 0));
END component;

signal 	high 		: std_logic ;
signal low 		: std_logic ;
signal oe_int 		: std_logic ;
signal we_int 		: std_logic ;
signal low1 		: std_logic_vector(0 downto 0) ;
signal low2 		: std_logic_vector(1 downto 0) ;
signal low4 		: std_logic_vector(3 downto 0) ;
signal low8 		: std_logic_vector(7 downto 0) ;
signal low32 		: std_logic_vector(31 downto 0) ;
signal douta 		: std_logic_vector(17 downto 0) ;
signal doutb 		: std_logic_vector(35 downto 0) ;
signal dina 		: std_logic_vector(17 downto 0) ;
signal ram_douta 	: std_logic_vector(15 downto 0) ;
signal ram_doutb 	: std_logic_vector(31 downto 0) ;
signal ram_pouta 	: std_logic_vector(1 downto 0) ;
signal ram_poutb 	: std_logic_vector(3 downto 0) ;
signal ram_pina 	: std_logic_vector(1 downto 0) ;
signal ram_dina 	: std_logic_vector(15 downto 0) ;
signal ram_addra 	: std_logic_vector(9 downto 0) ;
signal ram_addrb 	: std_logic_vector(8 downto 0) ;
signal next_wr_address 	: std_logic_vector((addrsize-1) downto 0) ;

begin

low <= '0' ;
low1 <= "0" ;
low2 <= "00" ;
low4 <= "0000" ;
low8 <= "00000000" ;
low32 <= (others => '0') ;
high <= '1' ;
oe_int <= oe or rst ; --needed to make sure the ram enable is high during reset (because its a synchronous reset)
we_int <= we or rst ; --needed to make sure the ram enable is high during reset (because its a synchronous reset)

douta <= ram_pouta & ram_douta ;
doutb <= ram_poutb(3 downto 2) & ram_doutb(31 downto 16) & ram_poutb(1 downto 0) & ram_doutb(15 downto 0) ;

process (next_wr_address, datain)
begin
dina <= (others => '0') ;
dina(17 downto (18-addrsize-datasize)) <= next_wr_address & datain ;
end process ;

ram_pina <= dina(17 downto 16) ;
ram_dina <= dina(15 downto 0) ;

process (douta)
begin
ram_addra <= (others => '0' ) ; -- write address
ram_addra((addrsize-1) downto 0) <= douta(17 downto (18-addrsize)) ;
end process ;

process (doutb)
begin
ram_addrb <= (others => '0' ) ; -- read address
ram_addrb((addrsize-2) downto 0) <= doutb(35 downto (37-addrsize)) ;
end process ;

dataout <= ram_doutb((33-addrsize) downto (34-addrsize-datasize)) & ram_doutb((17-addrsize) downto (18-addrsize-datasize)) ;

flag(2) <= '1' when ram_poutb(3 downto 2) = "00" and ram_pouta(1 downto 0) = "11" else
           '1' when ram_poutb(3 downto 2) = "01" and ram_pouta(1 downto 0) = "10" else
           '1' when ram_poutb(3 downto 2) = "11" and ram_pouta(1 downto 0) = "00" else
           '1' when ram_poutb(3 downto 2) = "10" and ram_pouta(1 downto 0) = "01" else
           '0' ;

flag(1) <= '1' when ram_poutb(3 downto 2) = "00" and ram_pouta(1 downto 0) = "01" else
           '1' when ram_poutb(3 downto 2) = "01" and ram_pouta(1 downto 0) = "11" else
           '1' when ram_poutb(3 downto 2) = "11" and ram_pouta(1 downto 0) = "10" else
           '1' when ram_poutb(3 downto 2) = "10" and ram_pouta(1 downto 0) = "00" else
           '0' ;


flag(0) <= '1' when ram_poutb(3 downto 2) = "00" and ram_pouta(1 downto 0) = "00" else
           '1' when ram_poutb(3 downto 2) = "01" and ram_pouta(1 downto 0) = "01" else
           '1' when ram_poutb(3 downto 2) = "11" and ram_pouta(1 downto 0) = "11" else
           '1' when ram_poutb(3 downto 2) = "10" and ram_pouta(1 downto 0) = "10" else
           '0' ;

block_ram : ramb16_s18_s36 port map (
	DIA    		=> ram_dina ,
	DIB    		=> low32,
	DIPA    	=> ram_pina,
	DIPB    	=> low4,
	ENA    		=> we_int,
	ENB    		=> oe_int,
	WEA    		=> we,
	WEB    		=> low,
	SSRA   		=> rst,
	SSRB   		=> rst,
	clkA   		=> wr_clk,
	clkB   		=> rd_clk,
	addra  		=> ram_addra,
	addrb  		=> ram_addrb,
	DOA    		=> ram_douta,
	DOB    		=> ram_doutb,
	DOPA    	=> ram_pouta,
	DOPB    	=> ram_poutb);

inc : incrementer
generic map (
	size		=> addrsize)
port map (
	ip 		=> douta(17 downto (18-addrsize)),
	op 		=> next_wr_address );

end arch_saf_1to2;

